/* ./mk_graph  w  d  r  output_file
   Generate a d-regular r-restricted grid graph of size w x w

   Output:
   v0 w0
   v1 w1
   ...
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

typedef unsigned int DATA ;
typedef int RDATA ;

#define MAX(A, B) A > B ? A : B
#define MIN(A, B) A < B ? A : B

#define __OK__ 1
#define __NG__ 0

#define __WITH_CONNECTED__

RDATA get_distance(DATA v, DATA w, DATA width)
{
    DATA vx,vy,wx,wy;
    vx = v / width; vy = v % width;
    wx = w / width; wy = w % width;
    int dx = vx - wx;
    int dy = vy - wy;
    if( dx < 0 ) dx *= -1;
    if( dy < 0 ) dy *= -1;
    return dx + dy;
}
int confirm_dist(DATA v, DATA w, DATA width, RDATA r)
{
    if( get_distance(v,w,width) <= r ) return __OK__;
    else return __NG__;
}
/* execute two-toggle operation */
void two_toggle_operation(DATA width, DATA d, RDATA r, DATA E, DATA list_E[])
{
    DATA e1, e2, e1_v, e1_w, e2_v, e2_w;
    DATA new_e1_v, new_e1_w, new_e2_v, new_e2_w;
    unsigned int round=0;
    while( round < 2*E ) {
        do{
            e1 = random() % E;
            e2 = random() % E;
        }while( e1 == e2 );
        e1_v = list_E[2*e1];  e1_w = list_E[2*e1+1];
        e2_v = list_E[2*e2];  e2_w = list_E[2*e2+1];
        if( (confirm_dist(e1_v,e2_v,width,r) == __OK__) &&
            (confirm_dist(e1_w,e2_w,width,r) == __OK__) ) {
            new_e1_v = e1_v;  new_e1_w = e2_v;
            new_e2_v = e1_w;  new_e2_w = e2_w;
        }else if( (confirm_dist(e1_v,e2_w,width,r) == __OK__) &&
                  (confirm_dist(e1_w,e2_v,width,r) == __OK__) ) {
            new_e1_v = e1_v;  new_e1_w = e2_w;
            new_e2_v = e1_w;  new_e2_w = e2_v;
        }else{
            continue;
        }
        list_E[2*e1] = new_e1_v;  list_E[2*e1+1] = new_e1_w;
        list_E[2*e2] = new_e2_v;  list_E[2*e2+1] = new_e2_w;
        round ++;
    }
}
/* check whether the graph is connected or not, using BFS
   If the graph is connected, return __OK__=1. Otherwise __NG__=0. */
int exec_bfs(DATA width, DATA d, RDATA r, DATA E, DATA list_E[])
{
    DATA i, k, v, w, next_v, root, time;
    DATA q_top, q_rear;
    DATA n=width*width;
    DATA *visited = malloc(sizeof(DATA)*n);
    DATA *adjlist = malloc(sizeof(DATA)*n*d);
    DATA *degree = malloc(sizeof(DATA)*n);
    DATA *q=malloc(sizeof(DATA)*n);

    for(i=0;i<n;i++) degree[i] = 0;
    for(i=0;i<E;i++) {
        v = MIN(list_E[2*i],list_E[2*i+1]);
        w = MAX(list_E[2*i],list_E[2*i+1]);
        adjlist[ v*d+degree[v] ] = w;
        degree[v] ++;
        adjlist[ w*d+degree[w] ] = v;
        degree[w] ++;
    }
    if( (n*d) % 2 == 1 ) adjlist[n*d-1] = n+1;
    /* execute BFS */
    q_top = q_rear = 0;
    q[0] = root = 0;
    visited[root] = time = 0;
    time++;
    for(i=1;i<n;i++) visited[i] = n+1;
    while( q_top <= q_rear ) {
        v = q[q_top];
        q_top++;
        for(k=0;k<d;k++) {
            next_v = adjlist[v*d+k];
            if( next_v == v ) continue;
            if( next_v >= n ) continue;
            if( visited[next_v] > n  ) {
                visited[next_v] = time;
                time++;
                q[q_rear] = next_v;
                q_rear ++;
            }
        }
    }
    free(visited);
    free(adjlist);
    free(degree);
    free(q);
    if( time == n ) return __OK__;
    else return __NG__;
}
int main(int argc, char *argv[])
{
    DATA width, n, d, E;
    DATA x, y, k, eid;
    RDATA r;
    DATA seed;
    FILE *fp;
	
	if (argc < 5) {
		printf("Usage: %s width degree reach filename [seed]\n", argv[0]);
		return 1;
	}
    width = (DATA) atoi(argv[1]);
    d = (DATA) atoi(argv[2]);
    r = (DATA) atoi(argv[3]);
    fp=fopen(argv[4],"w");
	if (argc >= 6) {
		seed=atoi(argv[5]);
	} else {
		seed=(unsigned)time(NULL);
	}
    srandom(seed);

    n = width * width;
    E = n * d / 2;
    DATA *list_E = malloc(sizeof(DATA)*2*E);
    printf("(n,d,r,seed)=(%d,%d,%d,%d)\n",n,d,r,seed);
    /* Generate an initial d-regular r-restricted grid graph */
    eid=0;
    for(y=0;y<width;y++) {
        for(x=0;x<width/2;x++) {
            for(k=0;k<d;k++) {
                list_E[2*eid]   = y*width + 2*x;
                list_E[2*eid+1] = y*width + 2*x+1;
                eid++;
            }
        }
    }
    if( width % 2 == 1 ) {
        for(y=0;y<width/2;y++) {
            for(k=0;k<d;k++) {
                list_E[2*eid]   = width*2*y + width-1;
                list_E[2*eid+1] = width*(2*y+1) + width-1;
                eid++;
            }
        }
        /* add self-loop */
        for(k=0;k<d/2;k++) {
            list_E[2*eid] = list_E[2*eid+1] = n-1;
            eid++;
        }
    }

    /* Execute two-toggle operation  */
    two_toggle_operation(width,d,r,E,list_E);

#ifdef __WITH_CONNECTED__
    /* Execute two-toggle operation until the graph is connected, 
       if __WITH_CONNECTED__ is defined. */
    while( exec_bfs(width,d,r,E,list_E) == __NG__ )
        two_toggle_operation(width,d,r,E,list_E);
#endif

    /* output edges */
    for(eid=0;eid<E;eid++) 
        fprintf(fp,"%u %u\n",list_E[2*eid],list_E[2*eid+1]);

    fclose(fp);
    return 0;
}
